#!/usr/bin/env python
# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that gets the sustainability data of a CPC.
"""

import sys
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

range = "last-day"
resolution = "fifteen-minutes"

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)
    format_str = "{:<8}  {:<6}  {:<7}  {:<16}"
    rc = 0

    cpcs = client.cpcs.list()
    cpc = cpcs[0]
    print('')
    print(f'Getting sustainability metrics on CPC: {cpc.name}')
    print(f'Range: {range}')
    print(f'Resolution: {resolution}')
    try:
        data = cpc.get_sustainability_data(
            range=range, resolution=resolution)
    except zhmcclient.Error as exc:
        print(f"Error: {exc}")
        rc = 1
    else:
        print('')
        print('CPC sustainability metrics:')
        for metric_name, metric_array in data.items():
            print(f"{metric_name}:")
            for dp in metric_array:
                print(f"  {dp['timestamp']}: {dp['data']}")

    if cpc.dpm_enabled:
        parts = cpc.partitions.list()
        part_str = "Partition"
    else:
        parts = cpc.lpars.list()
        part_str = "LPAR"
    part = parts[0]
    print('')
    print(f'Getting sustainability metrics on {part_str}: {part.name}')
    print(f'Range: {range}')
    print(f'Resolution: {resolution}')
    try:
        data = part.get_sustainability_data(
            range=range, resolution=resolution)
    except zhmcclient.Error as exc:
        print(f"Error: {exc}")
        rc = 1
    else:
        print('')
        print(f'{part_str} sustainability metrics:')
        for metric_name, metric_array in data.items():
            print(f"{metric_name}:")
            for dp in metric_array:
                print(f"  {dp['timestamp']}: {dp['data']}")

    if rc != 0:
        print("Error happened - see above")
        sys.exit(rc)

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _nic module.
"""


import re
import copy
import pytest

from zhmcclient import Client, Nic, Port, HTTPError, NotFound
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources


# Object IDs and names of our faked NICs:
NIC1_OID = 'nic 1-oid'
NIC1_NAME = 'nic 1'
NIC2_OID = 'nic 2-oid'
NIC2_NAME = 'nic 2'

# URIs and Object IDs of elements referenced in NIC properties:
VSWITCH11_OID = 'fake-vswitch11-oid'
VSWITCH11_URI = f'/api/virtual-switches/{VSWITCH11_OID}'
ROCE2_OID = 'fake-roce2-oid'
PORT21_OID = 'fake-port21-oid'
PORT21_URI = f'/api/adapters/{ROCE2_OID}/network-ports/{PORT21_OID}'


class TestNic:
    """All tests for Nic and NicManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked CPC in DPM mode with one
        partition that has no NICs.
        Add one OSA adapter, port and vswitch, for tests with OSA-backed NICs.
        Add one ROSE adapter and port, for tests with ROCE-backed NICs.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        # Add a CPC in DPM mode
        self.faked_cpc = self.session.hmc.cpcs.add({
            'element-id': 'fake-cpc1-oid',
            # element-uri is set up automatically
            'parent': None,
            'class': 'cpc',
            'name': 'fake-cpc1-name',
            'description': 'CPC #1 (DPM mode)',
            'status': 'active',
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'iml-mode': 'dpm',
        })
        self.cpc = self.client.cpcs.find(name='fake-cpc1-name')

        # Add a partition to the CPC
        self.faked_partition = self.faked_cpc.partitions.add({
            'element-id': 'fake-part1-oid',
            # element-uri will be automatically set
            'parent': self.faked_cpc.uri,
            'class': 'partition',
            'name': 'fake-part1-name',
            'description': 'Partition #1',
            'status': 'active',
            'initial-memory': 1024,
            'maximum-memory': 2048,
        })
        self.partition = self.cpc.partitions.find(name='fake-part1-name')

        # Add an OSA adapter, port and vswitch to the CPC
        self.faked_osa1 = self.faked_cpc.adapters.add({
            'object-id': 'osa1-oid',
            'parent': self.faked_cpc.uri,
            'class': 'adapter',
            'name': 'osa1',
            'description': 'OSA #1',
            'status': 'active',
            'type': 'osd',
            'adapter-id': '123',
            'detected-card-type': 'osa-express-5s-10gb',
            'card-location': '1234-5678-J.01',
            'port-count': 1,
            'network-port-uris': [],
            'state': 'online',
            'configured-capacity': 80,
            'used-capacity': 0,
            'allowed-capacity': 80,
            'maximum-total-capacity': 80,
            'channel-path-id': '1D',
            'physical-channel-status': 'operating',
        })
        self.osa1 = self.cpc.adapters.find(name='osa1')

        self.faked_port11 = self.faked_osa1.ports.add({
            'element-id': 'fake-port11-oid',
            'parent': self.faked_osa1.uri,
            'class': 'network-port',
            'index': 0,
            'name': 'fake-port11-name',
            'description': 'OSA #1 Port #1',
        })
        self.port11 = self.osa1.ports.find(name='fake-port11-name')

        self.faked_vswitch11 = self.faked_cpc.virtual_switches.add({
            'object-id': VSWITCH11_OID,
            'parent': self.faked_cpc.uri,
            'class': 'virtual-switch',
            'name': 'fake-vswitch11-name',
            'description': 'Vswitch for OSA #1 Port #1',
            'type': 'osa',
            'backing-adapter-uri': self.faked_osa1.uri,
            'port': self.faked_port11.properties['index'],
            'connected-vnic-uris': [],
        })
        self.vswitch11 = self.cpc.virtual_switches.find(
            name='fake-vswitch11-name')
        assert VSWITCH11_URI == self.faked_vswitch11.uri

        # Add a ROCE adapter and port to the CPC
        self.faked_roce2 = self.faked_cpc.adapters.add({
            'object-id': ROCE2_OID,
            'parent': self.faked_cpc.uri,
            'class': 'adapter',
            'name': 'roce2',
            'description': 'ROCE #2',
            'status': 'active',
            'type': 'roce',
            'adapter-id': '123',
            'detected-card-type': '10gbe-roce-express',
            'card-location': '1234-5678-J.01',
            'port-count': 1,
            'network-port-uris': [],
            'state': 'online',
            'configured-capacity': 80,
            'used-capacity': 0,
            'allowed-capacity': 80,
            'maximum-total-capacity': 80,
            'physical-channel-status': 'operating',
        })
        self.roce2 = self.cpc.adapters.find(name='roce2')

        self.faked_port21 = self.faked_roce2.ports.add({
            'element-id': PORT21_OID,
            'parent': self.faked_roce2.uri,
            'class': 'network-port',
            'index': 1,
            'name': 'fake-port21-name',
            'description': 'ROCE #2 Port #1',
        })
        self.port21 = self.roce2.ports.find(name='fake-port21-name')
        assert PORT21_URI == self.faked_port21.uri

    def add_nic1(self):
        """Add a faked OSA NIC 1 to the faked partition."""
        faked_nic = self.faked_partition.nics.add({
            'element-id': NIC1_OID,
            # element-uri will be automatically set
            'parent': self.faked_partition.uri,
            'class': 'nic',
            'name': NIC1_NAME,
            'description': 'NIC ' + NIC1_NAME,
            'type': 'osd',
            'virtual-switch-uri': VSWITCH11_URI,
            'device-number': '1111',
            'ssc-management-nic': False,
        })
        return faked_nic

    def add_nic2(self):
        """Add a faked ROCE NIC 2 to the faked partition."""
        faked_nic = self.faked_partition.nics.add({
            'element-id': NIC2_OID,
            # element-uri will be automatically set
            'parent': self.faked_partition.uri,
            'class': 'nic',
            'name': NIC2_NAME,
            'description': 'NIC ' + NIC2_NAME,
            'type': 'roce',
            'network-adapter-port-uri': PORT21_URI,
            'device-number': '1112',
            'ssc-management-nic': False,
        })
        return faked_nic

    def test_nicmanager_initial_attrs(self):
        """Test initial attributes of NicManager."""

        nic_mgr = self.partition.nics

        # Verify all public properties of the manager object
        assert nic_mgr.resource_class == Nic
        assert nic_mgr.session == self.session
        assert nic_mgr.parent == self.partition
        assert nic_mgr.partition == self.partition

    # TODO: Test for NicManager.__repr__()

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            ({},
             ['element-uri']),
            (dict(full_properties=False),
             ['element-uri']),
            (dict(full_properties=True),
             None),
        ]
    )
    def test_nicmanager_list_full_properties(
            self, full_properties_kwargs, prop_names):
        """Test NicManager.list() with full_properties."""

        # Add two faked NICs
        faked_nic1 = self.add_nic1()
        faked_nic2 = self.add_nic2()

        exp_faked_nics = [faked_nic1, faked_nic2]
        nic_mgr = self.partition.nics

        # Execute the code to be tested
        nics = nic_mgr.list(**full_properties_kwargs)

        assert_resources(nics, exp_faked_nics, prop_names)

    @pytest.mark.parametrize(
        "filter_args, exp_oids", [
            ({'element-id': NIC1_OID},
             [NIC1_OID]),
            ({'element-id': NIC2_OID},
             [NIC2_OID]),
            ({'element-id': [NIC1_OID, NIC2_OID]},
             [NIC1_OID, NIC2_OID]),
            ({'element-id': [NIC1_OID, NIC1_OID]},
             [NIC1_OID]),
            ({'element-id': NIC1_OID + 'foo'},
             []),
            ({'element-id': [NIC1_OID, NIC2_OID + 'foo']},
             [NIC1_OID]),
            ({'element-id': [NIC2_OID + 'foo', NIC1_OID]},
             [NIC1_OID]),
            ({'name': NIC1_NAME},
             [NIC1_OID]),
            ({'name': NIC2_NAME},
             [NIC2_OID]),
            ({'name': [NIC1_NAME, NIC2_NAME]},
             [NIC1_OID, NIC2_OID]),
            ({'name': NIC1_NAME + 'foo'},
             []),
            ({'name': [NIC1_NAME, NIC2_NAME + 'foo']},
             [NIC1_OID]),
            ({'name': [NIC2_NAME + 'foo', NIC1_NAME]},
             [NIC1_OID]),
            ({'name': [NIC1_NAME, NIC1_NAME]},
             [NIC1_OID]),
            ({'name': '.*nic 1'},
             [NIC1_OID]),
            ({'name': 'nic 1.*'},
             [NIC1_OID]),
            ({'name': 'nic .'},
             [NIC1_OID, NIC2_OID]),
            ({'name': '.ic 1'},
             [NIC1_OID]),
            ({'name': '.+'},
             [NIC1_OID, NIC2_OID]),
            ({'name': 'nic 1.+'},
             []),
            ({'name': '.+nic 1'},
             []),
            ({'name': NIC1_NAME,
              'element-id': NIC1_OID},
             [NIC1_OID]),
            ({'name': NIC1_NAME,
              'element-id': NIC1_OID + 'foo'},
             []),
            ({'name': NIC1_NAME + 'foo',
              'element-id': NIC1_OID},
             []),
            ({'name': NIC1_NAME + 'foo',
              'element-id': NIC1_OID + 'foo'},
             []),
        ]
    )
    def test_nicmanager_list_filter_args(self, filter_args, exp_oids):
        """Test NicManager.list() with filter_args."""

        # Add two faked NICs
        self.add_nic1()
        self.add_nic2()

        nic_mgr = self.partition.nics

        # Execute the code to be tested
        nics = nic_mgr.list(filter_args=filter_args)

        assert len(nics) == len(exp_oids)
        if exp_oids:
            oids = [nic.properties['element-id'] for nic in nics]
            assert set(oids) == set(exp_oids)

    @pytest.mark.parametrize(
        "initial_partition_status, exp_status_exc", [
            ('stopped', None),
            ('terminated', None),
            ('starting', HTTPError({'http-status': 409, 'reason': 1})),
            ('active', None),
            ('stopping', HTTPError({'http-status': 409, 'reason': 1})),
            ('degraded', None),
            ('reservation-error', None),
            ('paused', None),
        ]
    )
    @pytest.mark.parametrize(
        "input_props, exp_prop_names, exp_prop_exc", [
            ({},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-nic-x'},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'network-adapter-port-uri': PORT21_URI},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'virtual-switch-uri': VSWITCH11_URI},
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'name': 'fake-nic-x',
              'network-adapter-port-uri': PORT21_URI},
             ['element-uri', 'name', 'network-adapter-port-uri'],
             None),
            ({'name': 'fake-nic-x',
              'virtual-switch-uri': VSWITCH11_URI},
             ['element-uri', 'name', 'virtual-switch-uri'],
             None),
        ]
    )
    def test_nicmanager_create(
            self, input_props, exp_prop_names, exp_prop_exc,
            initial_partition_status, exp_status_exc):
        """Test NicManager.create()."""

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = initial_partition_status

        nic_mgr = self.partition.nics

        if exp_status_exc:
            exp_exc = exp_status_exc
        elif exp_prop_exc:
            exp_exc = exp_prop_exc
        else:
            exp_exc = None

        if exp_exc:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                nic = nic_mgr.create(properties=input_props)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            # Note: the Nic object returned by Nic.create() has
            # the input properties plus 'element-uri' plus 'element-id'.
            nic = nic_mgr.create(properties=input_props)

            # Check the resource for consistency within itself
            assert isinstance(nic, Nic)
            nic_name = nic.name
            exp_nic_name = nic.properties['name']
            assert nic_name == exp_nic_name
            nic_uri = nic.uri
            exp_nic_uri = nic.properties['element-uri']
            assert nic_uri == exp_nic_uri

            # Check the properties against the expected names and values
            for prop_name in exp_prop_names:
                assert prop_name in nic.properties
                if prop_name in input_props:
                    value = nic.properties[prop_name]
                    exp_value = input_props[prop_name]
                    assert value == exp_value

    def test_nic_repr(self):
        """Test Nic.__repr__()."""

        # Add a faked nic
        faked_nic = self.add_nic1()

        nic_mgr = self.partition.nics
        nic = nic_mgr.find(name=faked_nic.name)

        # Execute the code to be tested
        repr_str = repr(nic)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{nic.__class__.__name__}\s+at\s+'
            rf'0x{id(nic):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "initial_partition_status, exp_exc", [
            ('stopped', None),
            ('terminated', None),
            ('starting', HTTPError({'http-status': 409, 'reason': 1})),
            ('active', None),
            ('stopping', HTTPError({'http-status': 409, 'reason': 1})),
            ('degraded', None),
            ('reservation-error', None),
            ('paused', None),
        ]
    )
    def test_nic_delete(self, initial_partition_status, exp_exc):
        """Test Nic.delete()."""

        # Add a faked NIC to be tested and another one
        faked_nic = self.add_nic1()
        self.add_nic2()

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = initial_partition_status

        nic_mgr = self.partition.nics

        nic = nic_mgr.find(name=faked_nic.name)

        if exp_exc:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                nic.delete()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the NIC still exists
            nic_mgr.find(name=faked_nic.name)

        else:

            # Execute the code to be tested.
            nic.delete()

            # Check that the NIC no longer exists
            with pytest.raises(NotFound) as exc_info:
                nic_mgr.find(name=faked_nic.name)

    def test_nic_delete_create_same_name(self):
        """Test Nic.delete() followed by Nic.create() with same name."""

        # Add a faked NIC to be tested and another one
        faked_nic = self.add_nic1()
        nic_name = faked_nic.name
        self.add_nic2()

        # Construct the input properties for a third NIC with same name
        part3_props = copy.deepcopy(faked_nic.properties)
        part3_props['description'] = 'Third NIC'

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = 'stopped'  # deletable

        nic_mgr = self.partition.nics
        nic = nic_mgr.find(name=nic_name)

        # Execute the deletion code to be tested.
        nic.delete()

        # Check that the NIC no longer exists
        with pytest.raises(NotFound):
            nic_mgr.find(name=nic_name)

        # Execute the creation code to be tested.
        nic_mgr.create(part3_props)

        # Check that the NIC exists again under that name
        nic3 = nic_mgr.find(name=nic_name)
        description = nic3.get_property('description')
        assert description == 'Third NIC'

    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New NIC description'},
            {'device-number': 'FEDC',
             'description': 'New NIC description'},
        ]
    )
    def test_nic_update_properties(self, input_props):
        """Test Nic.update_properties()."""

        # Add a faked NIC
        faked_nic = self.add_nic1()

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = 'stopped'  # updatable

        nic_mgr = self.partition.nics
        nic = nic_mgr.find(name=faked_nic.name)

        nic.pull_full_properties()
        saved_properties = copy.deepcopy(nic.properties)

        # Execute the code to be tested
        nic.update_properties(properties=input_props)

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in nic.properties
            prop_value = nic.properties[prop_name]
            assert prop_value == exp_prop_value

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        nic.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in nic.properties
            prop_value = nic.properties[prop_name]
            assert prop_value == exp_prop_value

    def test_nic_update_name(self):
        """Test Nic.update_properties() with 'name' property."""

        # Add a faked NIC
        faked_nic = self.add_nic1()
        nic_name = faked_nic.name

        # Set the status of the faked partition
        self.faked_partition.properties['status'] = 'stopped'  # updatable

        nic_mgr = self.partition.nics
        nic = nic_mgr.find(name=nic_name)

        new_nic_name = "new-" + nic_name

        # Execute the code to be tested
        nic.update_properties(properties={'name': new_nic_name})

        # Verify that the resource is no longer found by its old name, using
        # list() (this does not use the name-to-URI cache).
        nics_list = nic_mgr.list(
            filter_args=dict(name=nic_name))
        assert len(nics_list) == 0

        # Verify that the resource is no longer found by its old name, using
        # find() (this uses the name-to-URI cache).
        with pytest.raises(NotFound):
            nic_mgr.find(name=nic_name)

        # Verify that the resource object already reflects the update, even
        # though it has not been refreshed yet.
        assert nic.properties['name'] == new_nic_name

        # Refresh the resource object and verify that it still reflects the
        # update.
        nic.pull_full_properties()
        assert nic.properties['name'] == new_nic_name

        # Verify that the resource can be found by its new name, using find()
        new_nic_find = nic_mgr.find(name=new_nic_name)
        assert new_nic_find.properties['name'] == new_nic_name

        # Verify that the resource can be found by its new name, using list()
        new_nics_list = nic_mgr.list(
            filter_args=dict(name=new_nic_name))
        assert len(new_nics_list) == 1
        new_nic_list = new_nics_list[0]
        assert new_nic_list.properties['name'] == new_nic_name

    def test_nicmanager_resource_object(self):
        """
        Test NicManager.resource_object().

        This test exists for historical reasons, and by now is covered by the
        test for BaseManager.resource_object().
        """

        nic_mgr = self.partition.nics
        nic_oid = 'fake-nic-id0711'

        # Execute the code to be tested
        nic = nic_mgr.resource_object(nic_oid)

        nic_uri = self.partition.uri + "/nics/" + nic_oid

        assert isinstance(nic, Nic)
        assert nic.uri == nic_uri
        assert nic.properties['element-uri'] == nic_uri
        assert nic.properties['element-id'] == nic_oid
        assert nic.properties['class'] == 'nic'
        assert nic.properties['parent'] == self.partition.uri

    def test_nic_backing_port_vswitch_based(self):
        """Test Nic.backing_port() for vswitch-based NICs."""

        # Add a faked nic
        faked_nic = self.add_nic1()
        nic = self.partition.nics.find(name=faked_nic.name)

        # Execute the code to be tested
        port = nic.backing_port()

        assert isinstance(port, Port)
        assert port.uri == self.port11.uri

        adapter = port.manager.parent
        assert adapter.uri == self.osa1.uri

    def test_nic_backing_port_port_based(self):
        """Test Nic.backing_port() for port-based NICs."""

        # Add a faked nic
        faked_nic = self.add_nic2()
        nic = self.partition.nics.find(name=faked_nic.name)

        # Execute the code to be tested
        port = nic.backing_port()

        assert isinstance(port, Port)
        assert port.uri == self.port21.uri

        adapter = port.manager.parent
        assert adapter.uri == self.roce2.uri



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=protected-access,attribute-defined-outside-init

"""
Unit tests for _hmc module of the zhmcclient_mock package.
"""


import re
from datetime import datetime
from dateutil import tz
import pytz
import pytest

from zhmcclient_mock._session import FakedSession
from zhmcclient_mock._hmc import \
    FakedBaseManager, FakedBaseResource, \
    FakedActivationProfileManager, FakedActivationProfile, \
    FakedAdapterManager, FakedAdapter, \
    FakedCpcManager, FakedCpc, \
    FakedHbaManager, FakedHba, \
    FakedLparManager, FakedLpar, \
    FakedNicManager, FakedNic, \
    FakedPartitionManager, FakedPartition, \
    FakedPortManager, FakedPort, \
    FakedVirtualFunctionManager, FakedVirtualFunction, \
    FakedVirtualSwitchManager, FakedVirtualSwitch, \
    FakedCapacityGroupManager, FakedCapacityGroup, \
    FakedMetricsContextManager, FakedMetricsContext, \
    FakedMetricGroupDefinition, FakedMetricObjectValues

from tests.common.utils import timestamp_aware


class TestFakedHmc:
    """All tests for the zhmcclient_mock.FakedHmc class."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC with attributes set but no child resourcs.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc

    def test_hmc_repr(self):
        """Test FakedHmc.__repr__()."""

        hmc = self.hmc

        # the function to be tested:
        repr_str = repr(hmc)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf"^{hmc.__class__.__name__}\s+at\s+"
            rf"0x{id(hmc):08x}\s+\(\\n.*", repr_str)

    def test_hmc_attrs(self):
        """Test FakedHmc attributes."""

        assert self.hmc.hmc_name == 'fake-hmc'
        assert self.hmc.hmc_version == '2.13.1'
        assert self.hmc.api_version == '1.8'
        assert isinstance(self.hmc.cpcs, FakedCpcManager)

        cpcs = self.hmc.cpcs.list()
        assert len(cpcs) == 0

    def test_hmc_1_cpc(self):
        """Test FakedHmc, adding and listing one CPC."""

        cpc1_in_props = {'name': 'cpc1'}

        # the function to be tested:
        cpc1 = self.hmc.cpcs.add(cpc1_in_props)

        cpc1_out_props = cpc1_in_props.copy()
        cpc1_out_props.update({
            'object-id': cpc1.oid,
            'object-uri': cpc1.uri,
            'class': 'cpc',
            'parent': None,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })

        # the function to be tested:
        cpcs = self.hmc.cpcs.list()

        assert len(cpcs) == 1
        assert cpcs[0] == cpc1

        assert isinstance(cpc1, FakedCpc)
        assert cpc1.properties == cpc1_out_props
        assert cpc1.manager == self.hmc.cpcs

    def test_hmc_2_cpcs(self):
        """Test FakedHmc, adding and listing two CPCs."""

        cpc1_in_props = {'name': 'cpc1'}

        # the function to be tested:
        cpc1 = self.hmc.cpcs.add(cpc1_in_props)

        cpc1_out_props = cpc1_in_props.copy()
        cpc1_out_props.update({
            'object-id': cpc1.oid,
            'object-uri': cpc1.uri,
            'class': 'cpc',
            'parent': None,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })

        cpc2_in_props = {'name': 'cpc2'}

        # the function to be tested:
        cpc2 = self.hmc.cpcs.add(cpc2_in_props)

        cpc2_out_props = cpc2_in_props.copy()
        cpc2_out_props.update({
            'object-id': cpc2.oid,
            'object-uri': cpc2.uri,
            'class': 'cpc',
            'parent': None,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })

        # the function to be tested:
        cpcs = self.hmc.cpcs.list()

        assert len(cpcs) == 2
        # We expect the order of addition to be maintained:
        assert cpcs[0] == cpc1
        assert cpcs[1] == cpc2

        assert isinstance(cpc1, FakedCpc)
        assert cpc1.properties == cpc1_out_props
        assert cpc1.manager == self.hmc.cpcs

        assert isinstance(cpc2, FakedCpc)
        assert cpc2.properties == cpc2_out_props
        assert cpc2.manager == self.hmc.cpcs

    def test_hmc_add_resources(self):
        """Test FakedHmc.add_resources()."""

        cpc1_in_props = {'name': 'cpc1'}
        adapter1_in_props = {'name': 'osa1', 'adapter-family': 'hipersockets'}
        port1_in_props = {'name': 'osa1_1'}

        rd = {
            'cpcs': [
                {
                    'properties': cpc1_in_props,
                    'adapters': [
                        {
                            'properties': adapter1_in_props,
                            'ports': [
                                {'properties': port1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }

        # the function to be tested:
        self.hmc.add_resources(rd)

        cpcs = self.hmc.cpcs.list()

        assert len(cpcs) == 1

        cpc1 = cpcs[0]
        cpc1_out_props = cpc1_in_props.copy()
        cpc1_out_props.update({
            'object-id': cpc1.oid,
            'object-uri': cpc1.uri,
            'class': 'cpc',
            'parent': None,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })
        assert isinstance(cpc1, FakedCpc)
        assert cpc1.properties == cpc1_out_props
        assert cpc1.manager == self.hmc.cpcs

        cpc1_adapters = cpc1.adapters.list()

        assert len(cpc1_adapters) == 1
        adapter1 = cpc1_adapters[0]

        adapter1_ports = adapter1.ports.list()

        assert len(adapter1_ports) == 1
        port1 = adapter1_ports[0]

        adapter1_out_props = adapter1_in_props.copy()
        adapter1_out_props.update({
            'object-id': adapter1.oid,
            'object-uri': adapter1.uri,
            'class': 'adapter',
            'parent': cpc1.uri,
            'status': 'active',
            'network-port-uris': [port1.uri],
        })
        assert isinstance(adapter1, FakedAdapter)
        assert adapter1.properties == adapter1_out_props
        assert adapter1.manager == cpc1.adapters

        port1_out_props = port1_in_props.copy()
        port1_out_props.update({
            'element-id': port1.oid,
            'element-uri': port1.uri,
            'class': 'network-port',
            'parent': adapter1.uri,
        })
        assert isinstance(port1, FakedPort)
        assert port1.properties == port1_out_props
        assert port1.manager == adapter1.ports

    def test_hmc_session_id(self):
        """Test FakedHmc.validate/add/remove_session_id()."""

        hmc = self.hmc

        session_id = 'valid_id'

        hmc.add_session_id(session_id)

        result = hmc.validate_session_id(session_id)
        assert result is True

        # Adding an already valid ID fails
        with pytest.raises(ValueError):
            hmc.add_session_id(session_id)

        hmc.remove_session_id(session_id)

        result = hmc.validate_session_id(session_id)
        assert result is False

        # Removing an invalid ID fails
        with pytest.raises(ValueError):
            hmc.remove_session_id(session_id)


class TestFakedBase:
    """All tests for the FakedBaseManager and FakedBaseResource classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has one managed CPC.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc

        self.cpc1_oid = '42-abc-543'
        self.cpc1_uri = f'/api/cpcs/{self.cpc1_oid}'

        self.cpc1_in_props = {
            # All properties that are otherwise defaulted (but with non-default
            # values), plus 'name'.
            'object-id': self.cpc1_oid,
            'object-uri': self.cpc1_uri,
            'class': 'cpc',
            'parent': None,
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'status': 'service',
            'name': 'cpc1',
        }
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                },
            ]
        }
        self.hmc.add_resources(rd)
        self.cpc_manager = self.hmc.cpcs
        self.cpc_resource = self.hmc.cpcs.list()[0]
        self.cpc1_out_props = self.cpc1_in_props.copy()

    def test_resource_repr(self):
        """Test FakedBaseResource.__repr__()."""

        resource = self.cpc_resource

        repr_str = repr(resource)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf"^{resource.__class__.__name__}\s+at\s+"
            rf"0x{id(resource):08x}\s+\(\\n.*", repr_str)

    def test_manager_repr(self):
        """Test FakedBaseManager.__repr__()."""

        manager = self.cpc_manager

        repr_str = repr(manager)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf"^{manager.__class__.__name__}\s+at\s+"
            rf"0x{id(manager):08x}\s+\(\\n.*", repr_str)

    def test_manager_attr(self):
        """Test FakedBaseManager attributes."""

        assert isinstance(self.cpc_manager, FakedBaseManager)

        assert self.cpc_manager.hmc == self.hmc
        assert self.cpc_manager.parent == self.hmc
        assert self.cpc_manager.resource_class == FakedCpc
        assert self.cpc_manager.base_uri == '/api/cpcs'
        assert self.cpc_manager.oid_prop == 'object-id'
        assert self.cpc_manager.uri_prop == 'object-uri'

    def test_resource_attr(self):
        """Test FakedBaseResource attributes."""

        assert isinstance(self.cpc_resource, FakedBaseResource)

        assert self.cpc_resource.manager == self.cpc_manager
        assert self.cpc_resource.properties == self.cpc1_out_props
        assert self.cpc_resource.oid == self.cpc1_out_props['object-id']
        assert self.cpc_resource.uri == self.cpc1_out_props['object-uri']


class TestFakedActivationProfile:
    """All tests for the FakedActivationProfileManager and
    FakedActivationProfile classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC in classic mode with some
        activation profiles.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.resetprofile1_in_props = {'name': 'resetprofile1'}
        self.imageprofile1_in_props = {'name': 'imageprofile1'}
        self.loadprofile1_in_props = {'name': 'loadprofile1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'reset_activation_profiles': [
                        {'properties': self.resetprofile1_in_props},
                    ],
                    'image_activation_profiles': [
                        {'properties': self.imageprofile1_in_props},
                    ],
                    'load_activation_profiles': [
                        {'properties': self.loadprofile1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedActivationProfileManager:
        self.hmc.add_resources(rd)

    def test_profiles_attr(self):
        """Test CPC '*_activation_profiles' attributes."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # Test reset activation profiles

        assert isinstance(cpc1.reset_activation_profiles,
                          FakedActivationProfileManager)
        assert cpc1.reset_activation_profiles.profile_type == 'reset'
        assert re.match(r'/api/cpcs/[^/]+/reset-activation-profiles',
                        cpc1.reset_activation_profiles.base_uri)

        # Test image activation profiles

        assert isinstance(cpc1.image_activation_profiles,
                          FakedActivationProfileManager)
        assert cpc1.image_activation_profiles.profile_type == 'image'
        assert re.match(r'/api/cpcs/[^/]+/image-activation-profiles',
                        cpc1.image_activation_profiles.base_uri)

        # Test load activation profiles

        assert isinstance(cpc1.load_activation_profiles,
                          FakedActivationProfileManager)
        assert cpc1.load_activation_profiles.profile_type == 'load'
        assert re.match(r'/api/cpcs/[^/]+/load-activation-profiles',
                        cpc1.load_activation_profiles.base_uri)

    def test_profiles_list(self):
        """Test list() of FakedActivationProfileManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # Test reset activation profiles

        resetprofiles = cpc1.reset_activation_profiles.list()

        assert len(resetprofiles) == 1
        resetprofile1 = resetprofiles[0]
        resetprofile1_out_props = self.resetprofile1_in_props.copy()
        resetprofile1_out_props.update({
            'name': resetprofile1.oid,
            'element-uri': resetprofile1.uri,
            'class': 'reset-activation-profile',
            'parent': cpc1.uri,
        })
        assert isinstance(resetprofile1, FakedActivationProfile)
        assert resetprofile1.properties == resetprofile1_out_props
        assert resetprofile1.manager == cpc1.reset_activation_profiles

        # Test image activation profiles

        imageprofiles = cpc1.image_activation_profiles.list()

        assert len(imageprofiles) == 1
        imageprofile1 = imageprofiles[0]
        imageprofile1_out_props = self.imageprofile1_in_props.copy()
        imageprofile1_out_props.update({
            'name': imageprofile1.oid,
            'element-uri': imageprofile1.uri,
            'class': 'image-activation-profile',
            'parent': cpc1.uri,
        })
        assert isinstance(imageprofile1, FakedActivationProfile)
        assert imageprofile1.properties == imageprofile1_out_props
        assert imageprofile1.manager == cpc1.image_activation_profiles

        # Test load activation profiles

        loadprofiles = cpc1.load_activation_profiles.list()

        assert len(loadprofiles) == 1
        loadprofile1 = loadprofiles[0]
        loadprofile1_out_props = self.loadprofile1_in_props.copy()
        loadprofile1_out_props.update({
            'name': loadprofile1.oid,
            'element-uri': loadprofile1.uri,
            'class': 'load-activation-profile',
            'parent': cpc1.uri,
        })
        assert isinstance(loadprofile1, FakedActivationProfile)
        assert loadprofile1.properties == loadprofile1_out_props
        assert loadprofile1.manager == cpc1.load_activation_profiles

    def test_profiles_add(self):
        """Test add() of FakedActivationProfileManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        resetprofiles = cpc1.reset_activation_profiles.list()
        assert len(resetprofiles) == 1

        resetprofile2_in_props = {'name': 'resetprofile2'}

        # the function to be tested:
        new_resetprofile = cpc1.reset_activation_profiles.add(
            resetprofile2_in_props)

        resetprofiles = cpc1.reset_activation_profiles.list()
        assert len(resetprofiles) == 2

        resetprofile2 = [p for p in resetprofiles
                         if p.properties['name'] ==  # noqa: W504
                         resetprofile2_in_props['name']][0]

        assert new_resetprofile.properties == resetprofile2.properties
        assert new_resetprofile.manager == resetprofile2.manager

        resetprofile2_out_props = resetprofile2_in_props.copy()
        resetprofile2_out_props.update({
            'name': resetprofile2.oid,
            'element-uri': resetprofile2.uri,
            'class': 'reset-activation-profile',
            'parent': cpc1.uri,
        })
        assert isinstance(resetprofile2, FakedActivationProfile)
        assert resetprofile2.properties == resetprofile2_out_props
        assert resetprofile2.manager == cpc1.reset_activation_profiles

        # Because we know that the image and load profile managers are of the
        # same class, we don't need to test them.

    def test_profiles_remove(self):
        """Test remove() of FakedActivationProfileManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        resetprofiles = cpc1.reset_activation_profiles.list()
        resetprofile1 = resetprofiles[0]
        assert len(resetprofiles) == 1

        # the function to be tested:
        cpc1.reset_activation_profiles.remove(resetprofile1.oid)

        resetprofiles = cpc1.reset_activation_profiles.list()
        assert len(resetprofiles) == 0

        # Because we know that the image and load profile managers are of the
        # same class, we don't need to test them.


class TestFakedAdapter:
    """All tests for the FakedAdapterManager and FakedAdapter classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with one ROCE adapter.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.adapter1_in_props = {'name': 'adapter1', 'type': 'roce'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'adapters': [
                        {'properties': self.adapter1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedAdapterManager:
        self.hmc.add_resources(rd)

    def test_adapter_repr(self):
        """Test FakedAdapter.__repr__()."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter = adapters[0]

        repr_str = repr(adapter)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf"^{adapter.__class__.__name__}\s+at\s+"
            rf"0x{id(adapter):08x}\s+\(\\n.*", repr_str)

    def test_adapters_attr(self):
        """Test CPC 'adapters' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        assert isinstance(cpc1.adapters, FakedAdapterManager)
        assert re.match(r'/api/adapters', cpc1.adapters.base_uri)

    def test_adapters_list(self):
        """Test list() of FakedAdapterManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        adapters = cpc1.adapters.list()

        assert len(adapters) == 1
        adapter1 = adapters[0]
        adapter1_out_props = self.adapter1_in_props.copy()
        adapter1_out_props.update({
            'object-id': adapter1.oid,
            'object-uri': adapter1.uri,
            'class': 'adapter',
            'parent': cpc1.uri,
            'status': 'active',
            'adapter-family': 'roce',
            'network-port-uris': [],
        })
        assert isinstance(adapter1, FakedAdapter)
        assert adapter1.properties == adapter1_out_props
        assert adapter1.manager == cpc1.adapters

        # Quick check of child resources:
        assert isinstance(adapter1.ports, FakedPortManager)

    def test_adapters_add(self):
        """Test add() of FakedAdapterManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        assert len(adapters) == 1

        adapter2_in_props = {'name': 'adapter2', 'adapter-family': 'ficon'}

        # the function to be tested:
        new_adapter = cpc1.adapters.add(
            adapter2_in_props)

        adapters = cpc1.adapters.list()
        assert len(adapters) == 2

        adapter2 = [a for a in adapters
                    if a.properties['name'] == adapter2_in_props['name']][0]

        assert new_adapter.properties == adapter2.properties
        assert new_adapter.manager == adapter2.manager

        adapter2_out_props = adapter2_in_props.copy()
        adapter2_out_props.update({
            'object-id': adapter2.oid,
            'object-uri': adapter2.uri,
            'class': 'adapter',
            'parent': cpc1.uri,
            'status': 'active',
            'storage-port-uris': [],
        })
        assert isinstance(adapter2, FakedAdapter)
        assert adapter2.properties == adapter2_out_props
        assert adapter2.manager == cpc1.adapters

    def test_adapters_remove(self):
        """Test remove() of FakedAdapterManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]
        assert len(adapters) == 1

        # the function to be tested:
        cpc1.adapters.remove(adapter1.oid)

        adapters = cpc1.adapters.list()
        assert len(adapters) == 0


class TestFakedCpc:
    """All tests for the FakedCpcManager and FakedCpc classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                },
            ]
        }
        # This already uses add() of FakedCpcManager:
        self.hmc.add_resources(rd)

    def test_cpc_repr(self):
        """Test FakedCpc.__repr__()."""

        cpcs = self.hmc.cpcs.list()
        cpc = cpcs[0]

        repr_str = repr(cpc)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf"^{cpc.__class__.__name__}\s+at\s+"
            rf"0x{id(cpc):08x}\s+\(\\n.*", repr_str)

    def test_cpcs_attr(self):
        """Test HMC 'cpcs' attribute."""

        assert isinstance(self.hmc.cpcs, FakedCpcManager)
        assert re.match(r'/api/cpcs', self.hmc.cpcs.base_uri)

    def test_cpcs_list(self):
        """Test list() of FakedCpcManager."""

        # the function to be tested:
        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        cpc1_out_props = self.cpc1_in_props.copy()
        cpc1_out_props.update({
            'object-id': cpc1.oid,
            'object-uri': cpc1.uri,
            'class': 'cpc',
            'parent': None,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })
        assert isinstance(cpc1, FakedCpc)
        assert cpc1.properties == cpc1_out_props
        assert cpc1.manager == self.hmc.cpcs

        # Quick check of child resources:
        assert isinstance(cpc1.lpars, FakedLparManager)
        assert isinstance(cpc1.partitions, FakedPartitionManager)
        assert isinstance(cpc1.adapters, FakedAdapterManager)
        assert isinstance(cpc1.virtual_switches, FakedVirtualSwitchManager)
        assert isinstance(cpc1.capacity_groups, FakedCapacityGroupManager)
        assert isinstance(cpc1.reset_activation_profiles,
                          FakedActivationProfileManager)
        assert isinstance(cpc1.image_activation_profiles,
                          FakedActivationProfileManager)
        assert isinstance(cpc1.load_activation_profiles,
                          FakedActivationProfileManager)

    def test_cpcs_add(self):
        """Test add() of FakedCpcManager."""

        cpcs = self.hmc.cpcs.list()
        assert len(cpcs) == 1

        cpc2_in_props = {'name': 'cpc2'}

        # the function to be tested:
        new_cpc = self.hmc.cpcs.add(cpc2_in_props)

        cpcs = self.hmc.cpcs.list()
        assert len(cpcs) == 2

        cpc2 = [cpc for cpc in cpcs
                if cpc.properties['name'] == cpc2_in_props['name']][0]

        assert new_cpc.properties == cpc2.properties
        assert new_cpc.manager == cpc2.manager

        cpc2_out_props = cpc2_in_props.copy()
        cpc2_out_props.update({
            'object-id': cpc2.oid,
            'object-uri': cpc2.uri,
            'class': 'cpc',
            'parent': None,
            'dpm-enabled': False,
            'is-ensemble-member': False,
            'status': 'operating',
        })
        assert isinstance(cpc2, FakedCpc)
        assert cpc2.properties == cpc2_out_props
        assert cpc2.manager == self.hmc.cpcs

    def test_cpcs_remove(self):
        """Test remove() of FakedCpcManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        assert len(cpcs) == 1

        # the function to be tested:
        self.hmc.cpcs.remove(cpc1.oid)

        cpcs = self.hmc.cpcs.list()
        assert len(cpcs) == 0


class TestFakedHba:
    """All tests for the FakedHbaManager and FakedHba classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with an FCP adapter, and a partition
        that has an HBA.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc

        self.adapter1_oid = '747-abc-12345'
        self.adapter1_uri = f'/api/adapters/{self.adapter1_oid}'
        self.port1_oid = '23'
        self.port1_uri = (f'/api/adapters/{self.adapter1_oid}/storage-ports/'
                          f'{self.port1_oid}')
        self.hba1_oid = '999-123-xyz'

        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        self.adapter1_in_props = {
            'object-id': self.adapter1_oid,
            'name': 'fcp1',
            'type': 'fcp',
        }
        self.port1_in_props = {
            'element-id': self.port1_oid,
            'name': 'port1',
        }
        self.hba1_in_props = {
            'element-id': self.hba1_oid,
            'name': 'hba1',
            'adapter-port-uri': self.port1_uri,
        }
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {
                            'properties': self.partition1_in_props,
                            'hbas': [
                                {'properties': self.hba1_in_props},
                            ],
                        },
                    ],
                    'adapters': [
                        {
                            'properties': self.adapter1_in_props,
                            'ports': [
                                {'properties': self.port1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }
        # This already uses add() of FakedHbaManager:
        self.hmc.add_resources(rd)

    def test_hbas_attr(self):
        """Test Partition 'hbas' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        assert isinstance(partition1.hbas, FakedHbaManager)
        assert re.match(r'/api/partitions/[^/]+/hbas',
                        partition1.hbas.base_uri)

    def test_hbas_list(self):
        """Test list() of FakedHbaManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        # the function to be tested:
        hbas = partition1.hbas.list()

        assert len(hbas) == 1
        hba1 = hbas[0]
        hba1_out_props = self.hba1_in_props.copy()
        hba1_out_props.update({
            'element-id': self.hba1_oid,
            'element-uri': hba1.uri,
            'class': 'hba',
            'parent': partition1.uri,
            'device-number': hba1.properties['device-number'],
            'wwpn': hba1.properties['wwpn'],
        })
        assert isinstance(hba1, FakedHba)
        assert hba1.properties == hba1_out_props
        assert hba1.manager == partition1.hbas

    def test_hbas_add(self):
        """Test add() of FakedHbaManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        hbas = partition1.hbas.list()
        assert len(hbas) == 1

        hba2_oid = '22-55-xy'
        port_uri = '/api/adapters/abc-123/storage-ports/42'

        hba2_in_props = {
            'element-id': hba2_oid,
            'name': 'hba2',
            'adapter-port-uri': port_uri,
            'device-number': '8001',
            'wwpn': 'AFFEAFFE00008001',
        }

        # the function to be tested:
        new_hba = partition1.hbas.add(
            hba2_in_props)

        hbas = partition1.hbas.list()
        assert len(hbas) == 2

        hba2 = [hba for hba in hbas
                if hba.properties['name'] == hba2_in_props['name']][0]

        assert new_hba.properties == hba2.properties
        assert new_hba.manager == hba2.manager

        hba2_out_props = hba2_in_props.copy()
        hba2_out_props.update({
            'element-id': hba2_oid,
            'element-uri': hba2.uri,
            'class': 'hba',
            'parent': partition1.uri,
        })
        assert isinstance(hba2, FakedHba)
        assert hba2.properties == hba2_out_props
        assert hba2.manager == partition1.hbas

    def test_hbas_remove(self):
        """Test remove() of FakedHbaManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        hbas = partition1.hbas.list()
        hba1 = hbas[0]
        assert len(hbas) == 1

        # the function to be tested:
        partition1.hbas.remove(hba1.oid)

        hbas = partition1.hbas.list()
        assert len(hbas) == 0

    # TODO: Add testcases for updating 'hba-uris' parent property


# TODO: Add unit tests for FakedCapacityGroup/Manager.


class TestFakedLpar:
    """All tests for the FakedLparManager and FakedLpar classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with an LPAR.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.lpar1_in_props = {'name': 'lpar1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'lpars': [
                        {'properties': self.lpar1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedLparManager:
        self.hmc.add_resources(rd)

    def test_lpars_attr(self):
        """Test CPC 'lpars' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        assert isinstance(cpc1.lpars, FakedLparManager)
        assert re.match(r'/api/logical-partitions', cpc1.lpars.base_uri)

    def test_lpars_list(self):
        """Test list() of FakedLparManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        lpars = cpc1.lpars.list()

        assert len(lpars) == 1
        lpar1 = lpars[0]
        lpar1_out_props = self.lpar1_in_props.copy()
        lpar1_out_props.update({
            'object-id': lpar1.oid,
            'object-uri': lpar1.uri,
            'class': 'logical-partition',
            'parent': cpc1.uri,
            'status': 'not-activated',
        })
        assert isinstance(lpar1, FakedLpar)
        assert lpar1.properties == lpar1_out_props
        assert lpar1.manager == cpc1.lpars

    def test_lpars_add(self):
        """Test add() of FakedLparManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        lpars = cpc1.lpars.list()
        assert len(lpars) == 1

        lpar2_in_props = {'name': 'lpar2'}

        # the function to be tested:
        new_lpar = cpc1.lpars.add(
            lpar2_in_props)

        lpars = cpc1.lpars.list()
        assert len(lpars) == 2

        lpar2 = [p for p in lpars
                 if p.properties['name'] == lpar2_in_props['name']][0]

        assert new_lpar.properties == lpar2.properties
        assert new_lpar.manager == lpar2.manager

        lpar2_out_props = lpar2_in_props.copy()
        lpar2_out_props.update({
            'object-id': lpar2.oid,
            'object-uri': lpar2.uri,
            'class': 'logical-partition',
            'parent': cpc1.uri,
            'status': 'not-activated',
        })
        assert isinstance(lpar2, FakedLpar)
        assert lpar2.properties == lpar2_out_props
        assert lpar2.manager == cpc1.lpars

    def test_lpars_remove(self):
        """Test remove() of FakedLparManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        lpars = cpc1.lpars.list()
        lpar1 = lpars[0]
        assert len(lpars) == 1

        # the function to be tested:
        cpc1.lpars.remove(lpar1.oid)

        lpars = cpc1.lpars.list()
        assert len(lpars) == 0


class TestFakedNic:
    """All tests for the FakedNicManager and FakedNic classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with a ROCE adapter and one
        partition that has a NIC.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc

        self.adapter1_oid = '380-xyz-12345'
        self.adapter1_uri = f'/api/adapters/{self.adapter1_oid}'
        self.port1_oid = '32'
        self.port1_uri = (f'/api/adapters/{self.adapter1_oid}/'
                          f'network-ports/{self.port1_oid}')
        self.nic1_oid = 'ddd-999-123'

        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        self.nic1_in_props = {
            'element-id': self.nic1_oid,
            'name': 'nic1',
            'network-adapter-port-uri': self.port1_uri,
        }
        self.adapter1_in_props = {
            'object-id': self.adapter1_oid,
            'name': 'roce1',
            'type': 'roce',
        }
        self.port1_in_props = {
            'element-id': self.port1_oid,
            'name': 'port1',
        }
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {
                            'properties': self.partition1_in_props,
                            'nics': [
                                {'properties': self.nic1_in_props},
                            ],
                        },
                    ],
                    'adapters': [
                        {
                            'properties': self.adapter1_in_props,
                            'ports': [
                                {'properties': self.port1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }
        # This already uses add() of FakedNicManager:
        self.hmc.add_resources(rd)

    def test_nics_attr(self):
        """Test Partition 'nics' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        assert isinstance(partition1.nics, FakedNicManager)
        assert re.match(r'/api/partitions/[^/]+/nics',
                        partition1.nics.base_uri)

    def test_nics_list(self):
        """Test list() of FakedNicManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        # the function to be tested:
        nics = partition1.nics.list()

        assert len(nics) == 1
        nic1 = nics[0]
        nic1_out_props = self.nic1_in_props.copy()
        nic1_out_props.update({
            'element-id': self.nic1_oid,
            'element-uri': nic1.uri,
            'class': 'nic',
            'parent': partition1.uri,
            'device-number': nic1.properties['device-number'],
            'ssc-management-nic': False,
            'type': 'iqd',
        })
        assert isinstance(nic1, FakedNic)
        assert nic1.properties == nic1_out_props
        assert nic1.manager == partition1.nics

    def test_nics_add(self):
        """Test add() of FakedNicManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        nics = partition1.nics.list()
        assert len(nics) == 1

        nic2_oid = '77-55-ab'
        port_uri = '/api/adapters/abc-123/network-ports/42'

        nic2_in_props = {
            'element-id': nic2_oid,
            'name': 'nic2',
            'network-adapter-port-uri': port_uri,
        }

        # the function to be tested:
        new_nic = partition1.nics.add(
            nic2_in_props)

        nics = partition1.nics.list()
        assert len(nics) == 2

        nic2 = [nic for nic in nics
                if nic.properties['name'] == nic2_in_props['name']][0]

        assert new_nic.properties == nic2.properties
        assert new_nic.manager == nic2.manager

        nic2_out_props = nic2_in_props.copy()
        nic2_out_props.update({
            'element-id': nic2_oid,
            'element-uri': nic2.uri,
            'class': 'nic',
            'parent': partition1.uri,
            'device-number': nic2.properties['device-number'],
            'ssc-management-nic': False,
            'type': 'iqd',
        })
        assert isinstance(nic2, FakedNic)
        assert nic2.properties == nic2_out_props
        assert nic2.manager == partition1.nics

    def test_nics_remove(self):
        """Test remove() of FakedNicManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        nics = partition1.nics.list()
        nic1 = nics[0]
        assert len(nics) == 1

        # the function to be tested:
        partition1.nics.remove(nic1.oid)

        nics = partition1.nics.list()
        assert len(nics) == 0

    # TODO: Add testcases for updating 'nic-uris' parent property


class TestFakedPartition:
    """All tests for the FakedPartitionManager and FakedPartition classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with a partition.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {'properties': self.partition1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedPartitionManager:
        self.hmc.add_resources(rd)

    def test_partition_repr(self):
        """Test FakedPartition.__repr__()."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition = partitions[0]

        repr_str = repr(partition)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf"^{partition.__class__.__name__}\s+at\s+"
            rf"0x{id(partition):08x}\s+\(\\n.*", repr_str)

    def test_partitions_attr(self):
        """Test CPC 'partitions' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        assert isinstance(cpc1.partitions, FakedPartitionManager)
        assert re.match(r'/api/partitions', cpc1.partitions.base_uri)

    def test_partitions_list(self):
        """Test list() of FakedPartitionManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        partitions = cpc1.partitions.list()

        assert len(partitions) == 1
        partition1 = partitions[0]
        partition1_out_props = self.partition1_in_props.copy()
        partition1_out_props.update({
            'object-id': partition1.oid,
            'object-uri': partition1.uri,
            'class': 'partition',
            'parent': cpc1.uri,
            'status': 'stopped',
            'hba-uris': [],
            'nic-uris': [],
            'virtual-function-uris': [],
            'partition-link-uris': [],
            'storage-group-uris': [],
            'tape-link-uris': [],
        })
        assert isinstance(partition1, FakedPartition)
        assert partition1.properties == partition1_out_props
        assert partition1.manager == cpc1.partitions

    def test_partitions_add(self):
        """Test add() of FakedPartitionManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        assert len(partitions) == 1

        partition2_in_props = {'name': 'partition2'}

        # the function to be tested:
        new_partition = cpc1.partitions.add(
            partition2_in_props)

        partitions = cpc1.partitions.list()
        assert len(partitions) == 2

        partition2 = [p for p in partitions
                      if p.properties['name'] ==  # noqa: W504
                      partition2_in_props['name']][0]

        assert new_partition.properties == partition2.properties
        assert new_partition.manager == partition2.manager

        partition2_out_props = partition2_in_props.copy()
        partition2_out_props.update({
            'object-id': partition2.oid,
            'object-uri': partition2.uri,
            'class': 'partition',
            'parent': cpc1.uri,
            'status': 'stopped',
            'hba-uris': [],
            'nic-uris': [],
            'virtual-function-uris': [],
            'partition-link-uris': [],
            'storage-group-uris': [],
            'tape-link-uris': [],
        })
        assert isinstance(partition2, FakedPartition)
        assert partition2.properties == partition2_out_props
        assert partition2.manager == cpc1.partitions

    def test_partitions_remove(self):
        """Test remove() of FakedPartitionManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        assert len(partitions) == 1

        # the function to be tested:
        cpc1.partitions.remove(partition1.oid)

        partitions = cpc1.partitions.list()
        assert len(partitions) == 0


class TestFakedPort:
    """All tests for the FakedPortManager and FakedPort classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with an OSA adapter with one Port.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.adapter1_in_props = {'name': 'adapter1', 'adapter-family': 'osa'}
        self.port1_in_props = {'name': 'port1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'adapters': [
                        {
                            'properties': self.adapter1_in_props,
                            'ports': [
                                {'properties': self.port1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }
        # This already uses add() of FakedPortManager:
        self.hmc.add_resources(rd)

    def test_ports_attr(self):
        """Test Adapter 'ports' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]

        assert isinstance(adapter1.ports, FakedPortManager)
        assert re.match(r'/api/adapters/[^/]+/network-ports',
                        adapter1.ports.base_uri)

    def test_ports_list(self):
        """Test list() of FakedPortManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]

        # the function to be tested:
        ports = adapter1.ports.list()

        assert len(ports) == 1
        port1 = ports[0]
        port1_out_props = self.port1_in_props.copy()
        port1_out_props.update({
            'element-id': port1.oid,
            'element-uri': port1.uri,
            'class': 'network-port',
            'parent': adapter1.uri,
        })
        assert isinstance(port1, FakedPort)
        assert port1.properties == port1_out_props
        assert port1.manager == adapter1.ports

    def test_ports_add(self):
        """Test add() of FakedPortManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]
        ports = adapter1.ports.list()
        assert len(ports) == 1

        port2_in_props = {'name': 'port2'}

        # the function to be tested:
        new_port = adapter1.ports.add(
            port2_in_props)

        ports = adapter1.ports.list()
        assert len(ports) == 2

        port2 = [p for p in ports
                 if p.properties['name'] == port2_in_props['name']][0]

        assert new_port.properties == port2.properties
        assert new_port.manager == port2.manager

        port2_out_props = port2_in_props.copy()
        port2_out_props.update({
            'element-id': port2.oid,
            'element-uri': port2.uri,
            'class': 'network-port',
            'parent': adapter1.uri,
        })
        assert isinstance(port2, FakedPort)
        assert port2.properties == port2_out_props
        assert port2.manager == adapter1.ports

    def test_ports_remove(self):
        """Test remove() of FakedPortManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        adapters = cpc1.adapters.list()
        adapter1 = adapters[0]
        ports = adapter1.ports.list()
        port1 = ports[0]
        assert len(ports) == 1

        # the function to be tested:
        adapter1.ports.remove(port1.oid)

        ports = adapter1.ports.list()
        assert len(ports) == 0

    # TODO: Add testcases for updating 'network-port-uris' and
    #       'storage-port-uris' parent properties


class TestFakedVirtualFunction:
    """All tests for the FakedVirtualFunctionManager and FakedVirtualFunction
    classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with a partition that has a virtual
        function.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        self.virtual_function1_in_props = {'name': 'virtual_function1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {
                            'properties': self.partition1_in_props,
                            'virtual_functions': [
                                {'properties':
                                 self.virtual_function1_in_props},
                            ],
                        },
                    ],
                },
            ]
        }
        # This already uses add() of FakedVirtualFunctionManager:
        self.hmc.add_resources(rd)

    def test_virtual_functions_attr(self):
        """Test Partition 'virtual_functions' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        assert isinstance(partition1.virtual_functions,
                          FakedVirtualFunctionManager)
        assert re.match(r'/api/partitions/[^/]+/virtual-functions',
                        partition1.virtual_functions.base_uri)

    def test_virtual_functions_list(self):
        """Test list() of FakedVirtualFunctionManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]

        # the function to be tested:
        virtual_functions = partition1.virtual_functions.list()

        assert len(virtual_functions) == 1
        virtual_function1 = virtual_functions[0]
        virtual_function1_out_props = self.virtual_function1_in_props.copy()
        virtual_function1_out_props.update({
            'element-id': virtual_function1.oid,
            'element-uri': virtual_function1.uri,
            'class': 'virtual-function',
            'parent': partition1.uri,
            'device-number': virtual_function1.properties['device-number'],
        })
        assert isinstance(virtual_function1, FakedVirtualFunction)
        assert virtual_function1.properties == virtual_function1_out_props
        assert virtual_function1.manager == partition1.virtual_functions

    def test_virtual_functions_add(self):
        """Test add() of FakedVirtualFunctionManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        virtual_functions = partition1.virtual_functions.list()
        assert len(virtual_functions) == 1

        virtual_function2_in_props = {'name': 'virtual_function2'}

        # the function to be tested:
        new_virtual_function = partition1.virtual_functions.add(
            virtual_function2_in_props)

        virtual_functions = partition1.virtual_functions.list()
        assert len(virtual_functions) == 2

        virtual_function2 = [vf for vf in virtual_functions
                             if vf.properties['name'] ==  # noqa: W504
                             virtual_function2_in_props['name']][0]

        assert new_virtual_function.properties == virtual_function2.properties
        assert new_virtual_function.manager == virtual_function2.manager

        virtual_function2_out_props = virtual_function2_in_props.copy()
        virtual_function2_out_props.update({
            'element-id': virtual_function2.oid,
            'element-uri': virtual_function2.uri,
            'class': 'virtual-function',
            'parent': partition1.uri,
            'device-number': virtual_function2.properties['device-number'],
        })
        assert isinstance(virtual_function2, FakedVirtualFunction)
        assert virtual_function2.properties == virtual_function2_out_props
        assert virtual_function2.manager == partition1.virtual_functions

    def test_virtual_functions_remove(self):
        """Test remove() of FakedVirtualFunctionManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        partitions = cpc1.partitions.list()
        partition1 = partitions[0]
        virtual_functions = partition1.virtual_functions.list()
        virtual_function1 = virtual_functions[0]
        assert len(virtual_functions) == 1

        # the function to be tested:
        partition1.virtual_functions.remove(virtual_function1.oid)

        virtual_functions = partition1.virtual_functions.list()
        assert len(virtual_functions) == 0

    # TODO: Add testcases for updating 'virtual-function-uris' parent property


class TestFakedVirtualSwitch:
    """All tests for the FakedVirtualSwitchManager and FakedVirtualSwitch
    classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with a virtual switch.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.virtual_switch1_in_props = {'name': 'virtual_switch1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'virtual_switches': [
                        {'properties': self.virtual_switch1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedVirtualSwitchManager:
        self.hmc.add_resources(rd)

    def test_virtual_switches_attr(self):
        """Test CPC 'virtual_switches' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        assert isinstance(cpc1.virtual_switches, FakedVirtualSwitchManager)
        assert re.match(r'/api/virtual-switches',
                        cpc1.virtual_switches.base_uri)

    def test_virtual_switches_list(self):
        """Test list() of FakedVirtualSwitchManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        virtual_switches = cpc1.virtual_switches.list()

        assert len(virtual_switches) == 1
        virtual_switch1 = virtual_switches[0]
        virtual_switch1_out_props = self.virtual_switch1_in_props.copy()
        virtual_switch1_out_props.update({
            'object-id': virtual_switch1.oid,
            'object-uri': virtual_switch1.uri,
            'class': 'virtual-switch',
            'parent': cpc1.uri,
            'connected-vnic-uris': [],
        })
        assert isinstance(virtual_switch1, FakedVirtualSwitch)
        assert virtual_switch1.properties == virtual_switch1_out_props
        assert virtual_switch1.manager == cpc1.virtual_switches

    def test_virtual_switches_add(self):
        """Test add() of FakedVirtualSwitchManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        virtual_switches = cpc1.virtual_switches.list()
        assert len(virtual_switches) == 1

        virtual_switch2_in_props = {'name': 'virtual_switch2'}

        # the function to be tested:
        new_virtual_switch = cpc1.virtual_switches.add(
            virtual_switch2_in_props)

        virtual_switches = cpc1.virtual_switches.list()
        assert len(virtual_switches) == 2

        virtual_switch2 = [p for p in virtual_switches
                           if p.properties['name'] ==  # noqa: W504
                           virtual_switch2_in_props['name']][0]

        assert new_virtual_switch.properties == virtual_switch2.properties
        assert new_virtual_switch.manager == virtual_switch2.manager

        virtual_switch2_out_props = virtual_switch2_in_props.copy()
        virtual_switch2_out_props.update({
            'object-id': virtual_switch2.oid,
            'object-uri': virtual_switch2.uri,
            'class': 'virtual-switch',
            'parent': cpc1.uri,
            'connected-vnic-uris': [],
        })
        assert isinstance(virtual_switch2, FakedVirtualSwitch)
        assert virtual_switch2.properties == virtual_switch2_out_props
        assert virtual_switch2.manager == cpc1.virtual_switches

    def test_virtual_switches_remove(self):
        """Test remove() of FakedVirtualSwitchManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        virtual_switches = cpc1.virtual_switches.list()
        virtual_switch1 = virtual_switches[0]
        assert len(virtual_switches) == 1

        # the function to be tested:
        cpc1.virtual_switches.remove(virtual_switch1.oid)

        virtual_switches = cpc1.virtual_switches.list()
        assert len(virtual_switches) == 0


class TestFakedCapacityGroup:
    """All tests for the FakedCapacityGroupManager and FakedCapacityGroup
    classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with a capacity group.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.capacity_group1_in_props = {'name': 'capacity_group1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'capacity_groups': [
                        {'properties': self.capacity_group1_in_props},
                    ],
                },
            ]
        }
        # This already uses add() of FakedCapacityGroupManager:
        self.hmc.add_resources(rd)

    def test_capacity_groups_attr(self):
        """Test CPC 'capacity_groups' attribute."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        assert isinstance(cpc1.capacity_groups, FakedCapacityGroupManager)
        assert re.match(cpc1.uri + '/capacity-groups',
                        cpc1.capacity_groups.base_uri)

    def test_capacity_groups_list(self):
        """Test list() of FakedCapacityGroupManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]

        # the function to be tested:
        capacity_groups = cpc1.capacity_groups.list()

        assert len(capacity_groups) == 1
        capacity_group1 = capacity_groups[0]
        capacity_group1_out_props = self.capacity_group1_in_props.copy()
        capacity_group1_out_props.update({
            'element-id': capacity_group1.oid,
            'element-uri': capacity_group1.uri,
            'class': 'capacity-group',
            'parent': cpc1.uri,
            'partition-uris': [],
            'capping-enabled': True,
        })
        assert isinstance(capacity_group1, FakedCapacityGroup)
        assert capacity_group1.properties == capacity_group1_out_props
        assert capacity_group1.manager == cpc1.capacity_groups

    def test_capacity_groups_add(self):
        """Test add() of FakedCapacityGroupManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        capacity_groups = cpc1.capacity_groups.list()
        assert len(capacity_groups) == 1

        capacity_group2_in_props = {'name': 'capacity_group2'}

        # the function to be tested:
        new_capacity_group = cpc1.capacity_groups.add(
            capacity_group2_in_props)

        capacity_groups = cpc1.capacity_groups.list()
        assert len(capacity_groups) == 2

        capacity_group2 = [p for p in capacity_groups
                           if p.properties['name'] ==  # noqa: W504
                           capacity_group2_in_props['name']][0]

        assert new_capacity_group.properties == capacity_group2.properties
        assert new_capacity_group.manager == capacity_group2.manager

        capacity_group2_out_props = capacity_group2_in_props.copy()
        capacity_group2_out_props.update({
            'element-id': capacity_group2.oid,
            'element-uri': capacity_group2.uri,
            'class': 'capacity-group',
            'parent': cpc1.uri,
            'partition-uris': [],
            'capping-enabled': True,
        })
        assert isinstance(capacity_group2, FakedCapacityGroup)
        assert capacity_group2.properties == capacity_group2_out_props
        assert capacity_group2.manager == cpc1.capacity_groups

    def test_capacity_groups_remove(self):
        """Test remove() of FakedCapacityGroupManager."""

        cpcs = self.hmc.cpcs.list()
        cpc1 = cpcs[0]
        capacity_groups = cpc1.capacity_groups.list()
        capacity_group1 = capacity_groups[0]
        assert len(capacity_groups) == 1

        # the function to be tested:
        cpc1.capacity_groups.remove(capacity_group1.oid)

        capacity_groups = cpc1.capacity_groups.list()
        assert len(capacity_groups) == 0


class TestFakedMetricsContext:
    """All tests for the FakedMetricsContextManager and FakedMetricsContext
    classes."""

    def setup_method(self):
        """
        Called by pytest before each test method.

        Creates a faked HMC that has a CPC with a partition.
        """
        session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.hmc = session.hmc
        self.cpc1_in_props = {'name': 'cpc1'}
        self.partition1_in_props = {'name': 'partition1'}
        rd = {
            'cpcs': [
                {
                    'properties': self.cpc1_in_props,
                    'partitions': [
                        {
                            'properties': self.partition1_in_props,
                        },
                    ],
                },
            ]
        }
        self.hmc.add_resources(rd)

    def test_mcm_attr(self):
        """Test faked HMC 'metrics_contexts' attribute."""

        faked_hmc = self.hmc

        assert isinstance(faked_hmc.metrics_contexts,
                          FakedMetricsContextManager)
        assert re.match(r'/api/services/metrics/context',
                        faked_hmc.metrics_contexts.base_uri)

    def test_mcm_add(self):
        """Test add() of FakedMetricsContextManager."""

        faked_hmc = self.hmc

        mc_in_props = {
            'anticipated-frequency-seconds': 1,
            'metric-groups': ['partition-usage'],
        }

        # the function to be tested:
        mc = faked_hmc.metrics_contexts.add(mc_in_props)

        assert isinstance(mc, FakedMetricsContext)
        assert re.match(r'/api/services/metrics/context/[^/]+', mc.uri)
        assert mc.manager is faked_hmc.metrics_contexts
        mc_props = mc_in_props.copy()
        mc_props.update({
            'fake-id': mc.oid,
            'fake-uri': mc.uri,
            'parent': None,
        })
        assert mc.properties == mc_props

    def test_mcm_add_get_metric_values(self):
        """
        Test add_metric_values() of FakedMetricsContextManager.
        """

        faked_hmc = self.hmc

        mg_name = 'partition-usage'
        mo_val_input = FakedMetricObjectValues(
            group_name=mg_name,
            resource_uri='/api/partitions/fake-oid',
            timestamp=datetime.now(),
            values=[
                ('metric-1', "a"),
                ('metric-2', 5),
            ])
        mo_val2_input = FakedMetricObjectValues(
            group_name=mg_name,
            resource_uri='/api/partitions/fake-oid2',
            timestamp=datetime.now(),
            values=[
                ('metric-1', "b"),
                ('metric-2', 7),
            ])

        # Verify the initial M.O.Val group names
        mo_val_group_names = faked_hmc.metric_values.keys()
        assert list(mo_val_group_names) == []

        # Verify that a first M.O.Val can be added
        faked_hmc.add_metric_values(mo_val_input)

        # Verify the M.O.Val group names after having added one
        mo_val_group_names = faked_hmc.metric_values.keys()
        assert list(mo_val_group_names) == [mg_name]

        # Verify that the M.O.Vals can be retrieved and contain the first one
        mo_vals = faked_hmc.metric_values[mg_name]
        assert list(mo_vals) == [mo_val_input]

        # Verify that a second M.O.Val can be added for the same group name
        faked_hmc.add_metric_values(mo_val2_input)

        # Verify the M.O.Val group names after having added a second M.O.Val
        # for the same group name -> still just one group name
        mo_val_group_names = faked_hmc.metric_values.keys()
        assert list(mo_val_group_names) == [mg_name]

        # Verify that the M.O.Vals can be retrieved and contain both
        mo_vals = faked_hmc.metric_values[mg_name]
        assert list(mo_vals) == [mo_val_input, mo_val2_input]

    def test_mc_get_mg_defs(self):
        """Test get_metric_group_definitions() of FakedMetricsContext."""

        faked_hmc = self.hmc

        mg_name = 'partition-usage'
        mg_def = faked_hmc.metric_groups[mg_name]
        mg_name2 = 'dpm-system-usage-overview'
        mg_def2 = faked_hmc.metric_groups[mg_name2]

        # Test case where only one M.G.Def is tested
        mc_in_props = {
            'anticipated-frequency-seconds': 1,
            'metric-groups': [mg_name],
        }
        mc = faked_hmc.metrics_contexts.add(mc_in_props)

        # the function to be tested:
        mg_defs = mc.get_metric_group_definitions()

        # Verify the returned M.G.Defs
        assert list(mg_defs) == [mg_def]

        # Test case where the default for M.G.Defs is tested
        mc_in_props = {
            'anticipated-frequency-seconds': 1,
            # 'metric-groups' not specified -> default: return all
        }
        mc = faked_hmc.metrics_contexts.add(mc_in_props)

        # the function to be tested:
        mg_defs = mc.get_metric_group_definitions()

        # Verify a subset of the returned M.G.Defs
        assert mg_def in mg_defs
        assert mg_def2 in mg_defs

    def test_mc_get_mg_infos(self):
        """Test get_metric_group_infos() of FakedMetricsContext."""

        faked_hmc = self.hmc

        mg_name = 'partition-usage'
        mg_def = faked_hmc.metric_groups[mg_name]
        mg_info = {
            'group-name': mg_name,
            'metric-infos': [
                {'metric-name': t[0], 'metric-type': t[1]}
                for t in mg_def.types
            ],
        }

        mg_name2 = 'dpm-system-usage-overview'
        mg_def2 = faked_hmc.metric_groups[mg_name2]
        mg_info2 = {
            'group-name': mg_name2,
            'metric-infos': [
                {'metric-name': t[0], 'metric-type': t[1]}
                for t in mg_def2.types
            ],
        }

        # Test case where only one M.G.Def is tested
        mc_in_props = {
            'anticipated-frequency-seconds': 1,
            'metric-groups': [mg_name],
        }
        mc = faked_hmc.metrics_contexts.add(mc_in_props)

        # the function to be tested:
        mg_infos = mc.get_metric_group_infos()

        # Verify the returned M.G.Defs
        assert list(mg_infos) == [mg_info]

        # Test case where the default for M.G.Defs is tested
        mc_in_props = {
            'anticipated-frequency-seconds': 1,
            # 'metric-groups' not specified -> default: return all
        }
        mc = faked_hmc.metrics_contexts.add(mc_in_props)

        # the function to be tested:
        mg_infos = mc.get_metric_group_infos()

        # Verify a subset of the returned M.G.Defs
        assert mg_info in mg_infos
        assert mg_info2 in mg_infos

    def test_mc_get_m_values(self):
        """Test get_metric_values() of FakedMetricsContext."""

        faked_hmc = self.hmc
        mc_mgr = faked_hmc.metrics_contexts

        mg_name = 'partition-usage'
        mo_val_input = FakedMetricObjectValues(
            group_name=mg_name,
            resource_uri='/api/partitions/fake-oid',
            timestamp=datetime.now(),
            values=[
                ('processor-usage', 10),
                ('network-usage', 5),
            ])
        faked_hmc.add_metric_values(mo_val_input)
        mo_val2_input = FakedMetricObjectValues(
            group_name=mg_name,
            resource_uri='/api/partitions/fake-oid2',
            timestamp=datetime.now(),
            values=[
                ('processor-usage', 12),
                ('network-usage', 3),
            ])
        faked_hmc.add_metric_values(mo_val2_input)

        exp_mo_vals = faked_hmc.metric_values[mg_name]

        # Test case where only one M.G.Def is tested
        mc_in_props = {
            'anticipated-frequency-seconds': 1,
            'metric-groups': [mg_name],
        }
        mc = mc_mgr.add(mc_in_props)

        # the function to be tested:
        mv_list = mc.get_metric_values()

        assert len(mv_list) == 1
        mv = mv_list[0]
        assert mv[0] == mg_name
        assert mv[1] == exp_mo_vals

    def test_mc_get_m_values_response(self):
        """Test get_metric_values_response() of FakedMetricsContext."""

        faked_hmc = self.hmc
        mc_mgr = faked_hmc.metrics_contexts

        mg_name = 'partition-usage'

        ts1_input = datetime(2017, 9, 5, 12, 13, 10, 0, pytz.utc)
        ts1_exp = 1504613590000
        mo_val_input = FakedMetricObjectValues(
            group_name=mg_name,
            resource_uri='/api/partitions/fake-oid',
            timestamp=ts1_input,
            values=[
                ('processor-usage', 12),
                ('network-usage', 3),
            ])
        faked_hmc.add_metric_values(mo_val_input)

        ts2_tz = pytz.timezone('CET')
        ts2_input = datetime(2017, 9, 5, 12, 13, 20, 0, ts2_tz)
        ts2_offset = int(ts2_tz.utcoffset(ts2_input).total_seconds())
        ts2_exp = 1504613600000 - 1000 * ts2_offset
        mo_val2_input = FakedMetricObjectValues(
            group_name=mg_name,
            resource_uri='/api/partitions/fake-oid',
            timestamp=ts2_input,
            values=[
                ('processor-usage', 10),
                ('network-usage', 5),
            ])
        faked_hmc.add_metric_values(mo_val2_input)

        mg_name2 = 'dpm-system-usage-overview'

        ts3_input = datetime(2017, 9, 5, 12, 13, 30, 0)  # timezone-naive
        ts3_offset = int(tz.tzlocal().utcoffset(ts3_input).total_seconds())
        ts3_exp = 1504613610000 - 1000 * ts3_offset
        mo_val3_input = FakedMetricObjectValues(
            group_name=mg_name2,
            resource_uri='/api/cpcs/fake-oid',
            timestamp=ts3_input,
            values=[
                ('processor-usage', 50),
                ('network-usage', 20),
            ])
        faked_hmc.add_metric_values(mo_val3_input)

        exp_mv_resp = f'''"partition-usage"
"/api/partitions/fake-oid"
{ts1_exp}
12,3

"/api/partitions/fake-oid"
{ts2_exp}
10,5


"dpm-system-usage-overview"
"/api/cpcs/fake-oid"
{ts3_exp}
50,20



'''

        # Test case where only one M.G.Def is tested
        mc_in_props = {
            'anticipated-frequency-seconds': 1,
            'metric-groups': [mg_name, mg_name2],
        }
        mc = mc_mgr.add(mc_in_props)

        # the function to be tested:
        mv_resp = mc.get_metric_values_response()

        assert mv_resp == exp_mv_resp, (
            f"Actual response string:\n{mv_resp!r}\n"
            f"Expected response string:\n{exp_mv_resp!r}\n")


def test_mgd_attr():
    """Test attributes of a FakedMetricGroupDefinition object."""

    in_kwargs = {
        'name': 'partition-usage',
        'types': [
            ('metric-1', 'string-metric'),
            ('metric-2', 'integer-metric'),
        ]
    }

    # the function to be tested:
    new_mgd = FakedMetricGroupDefinition(**in_kwargs)

    assert new_mgd.name == in_kwargs['name']
    assert new_mgd.types == in_kwargs['types']
    assert new_mgd.types is not in_kwargs['types']  # was copied


def test_mov_attr():
    """Test attributes of a FakedMetricObjectValues object."""

    in_timestamp = datetime.now()  # timezone-naive
    in_kwargs = {
        'group_name': 'partition-usage',
        'resource_uri': '/api/partitions/fake-oid',
        'timestamp': in_timestamp,
        'values': [
            ('metric-1', "a"),
            ('metric-2', 5),
        ]
    }
    exp_timestamp = timestamp_aware(in_timestamp)

    # the function to be tested:
    new_mov = FakedMetricObjectValues(**in_kwargs)

    assert new_mov.group_name == in_kwargs['group_name']
    assert new_mov.resource_uri == in_kwargs['resource_uri']
    assert new_mov.timestamp == exp_timestamp
    assert new_mov.values == in_kwargs['values']
    assert new_mov.values is not in_kwargs['values']  # was copied



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
HTTP-mocked pytest fixtures for unit tests.

HTTP-mocked pytest fixtures can be used in cases where the zhmcclient mock
support is not implemented, to mock at the level of the HTTP requests and
responses instead.

Example use in a pytest test function that invokes a (hypothetical) method
``Cpc.xyz()``:

.. code-block:: python

    from tests.common.http_mocked_fixtures import http_mocked_session
    from tests.common.http_mocked_fixtures import http_mocked_cpc_dpm


    def test_cpc_xyz(http_mocked_cpc_dpm):

        uri = http_mocked_cpc_dpm.uri + '/operations/xyz'

        rm_adapter = requests_mock.Adapter(case_sensitive=True)
        with requests_mock.mock(adapter=rm_adapter) as m:

            m.post(uri, status_code=200)

            result = http_mocked_cpc.xyz()

            assert rm_adapter.called
            request_body = rm_adapter.last_request.json()
            assert request_body == ...
            assert result == ...
"""

import pytest
import requests_mock
import zhmcclient


@pytest.fixture(
    scope='module'
)
def http_mocked_session(request):  # noqa: F811
    # pylint: disable=unused-argument
    """
    Pytest fixture representing a HTTP-mocked zhmcclient session that is logged
    on.

    A test function parameter using this fixture resolves to a
    :class:`~zhmcclient.Session` object that is logged on.
    """

    session = zhmcclient.Session(
        host='test-hmc', userid='test-user', password='test-pwd',
        verify_cert=False)

    with requests_mock.mock() as m:
        # Because logon is deferred until needed, we perform it
        # explicitly in order to keep mocking in the actual test simple.
        m.post('/api/sessions', status_code=200, json={
            'api-session': 'test-session-id.1',
            'notification-topic': 'test-obj-topic.1',
            'job-notification-topic': 'test-job-topic.1',
            'api-major-version': 4,
            'api-minor-version': 10,
            'password-expires': 90,
            'session-credential':
                'un8bu462g37aw9j0o8pltontz3szt35jh4b1qe2toxt6fkhl4'
        })
        m.get(
            '/api/version', json={
                'api-major-version': 4,
                'api-minor-version': 10,
            })
        session.logon()

    yield session

    with requests_mock.mock() as m:
        m.delete('/api/sessions/this-session', status_code=204)
        session.logoff()


@pytest.fixture(
    scope='module'
)
def http_mocked_cpc_dpm(request, http_mocked_session):  # noqa: F811
    # pylint: disable=unused-argument,redefined-outer-name
    """
    Pytest fixture representing a HTTP-mocked CPC in DPM mode.

    A test function parameter using this fixture resolves to a
    :class:`~zhmcclient.Cpc` object that is in DPM mode but has only a minimal
    set of properties in the object (those that are returned with the 'List
    CPCs' operation).
    """

    client = zhmcclient.Client(http_mocked_session)

    with requests_mock.mock() as m:
        m.get('/api/cpcs', status_code=200, json={
            'cpcs': [
                {
                    'object-uri': '/api/cpcs/cpc-id-1',
                    'name': 'CPC1',
                    'status': 'active',
                    'has-unacceptable-status': False,
                    'dpm-enabled': True,
                    'se-version': '2.16.0',
                }
            ]
        })
        cpcs = client.cpcs.list()
        cpc = cpcs[0]

    return cpc


@pytest.fixture(
    scope='module'
)
def http_mocked_cpc_classic(request, http_mocked_session):  # noqa: F811
    # pylint: disable=unused-argument,redefined-outer-name
    """
    Pytest fixture representing a HTTP-mocked CPC in classic mode.

    A test function parameter using this fixture resolves to a
    :class:`~zhmcclient.Cpc` object that is in classic mode but has only a
    minimal set of properties in the object (those that are returned with the
    'List CPCs' operation).
    """

    client = zhmcclient.Client(http_mocked_session)

    with requests_mock.mock() as m:
        m.get('/api/cpcs', status_code=200, json={
            'cpcs': [
                {
                    'object-uri': '/api/cpcs/cpc-id-2',
                    'name': 'CPC2',
                    'status': 'operating',
                    'has-unacceptable-status': False,
                    'dpm-enabled': False,
                    'se-version': '2.16.0',
                }
            ]
        })
        cpcs = client.cpcs.list()
        cpc = cpcs[0]

    return cpc


@pytest.fixture(
    scope='module'
)
def http_mocked_lpar(request, http_mocked_cpc_classic):  # noqa: F811
    # pylint: disable=unused-argument,redefined-outer-name
    """
    Pytest fixture representing a HTTP-mocked LPAR on a CPC in classic mode.

    Its CPC object can be accessed as the parent object.

    A test function parameter using this fixture resolves to a
    :class:`~zhmcclient.Lpar` object that has only a
    minimal set of properties in the object (those that are returned with the
    'List Logical Partitions of CPC' operation).
    """

    with requests_mock.mock() as m:
        uri = http_mocked_cpc_classic.uri + '/logical-partitions'
        m.get(uri, status_code=200, json={
            'logical-partitions': [
                {
                    'object-uri': '/api/logical-partitions/lpar-id-1',
                    'name': 'LPAR1',
                    'status': 'operating',
                }
            ]
        })
        lpars = http_mocked_cpc_classic.lpars.list()
        lpar = lpars[0]

    return lpar


@pytest.fixture(
    scope='module'
)
def http_mocked_partition(request, http_mocked_cpc_dpm):  # noqa: F811
    # pylint: disable=unused-argument,redefined-outer-name
    """
    Pytest fixture representing a HTTP-mocked Partition on a CPC in DPM mode.

    Its CPC object can be accessed as the parent object.

    A test function parameter using this fixture resolves to a
    :class:`~zhmcclient.Partition` object that has only a
    minimal set of properties in the object (those that are returned with the
    'List Partitions of CPC' operation).
    """

    with requests_mock.mock() as m:
        uri = http_mocked_cpc_dpm.uri + '/partitions'
        m.get(uri, status_code=200, json={
            'partitions': [
                {
                    'object-uri': '/api/partitions/part-id-1',
                    'name': 'PART1',
                    'status': 'active',
                }
            ]
        })
        partitions = http_mocked_cpc_dpm.partitions.list()
        partition = partitions[0]

    return partition


@pytest.fixture(
    scope='module'
)
def http_mocked_console(request, http_mocked_session):  # noqa: F811
    # pylint: disable=unused-argument,redefined-outer-name
    """
    Pytest fixture representing a HTTP-mocked HMC console.

    A test function parameter using this fixture resolves to a
    :class:`~zhmcclient.Console` object representing the HMC the session
    was opened against, that has only a subset of its properties in the object.
    """

    client = zhmcclient.Client(http_mocked_session)

    with requests_mock.mock() as m:
        m.get('/api/console', status_code=200, json={
            'object-uri': '/api/console',
            'parent': None,
            'class': 'console',
            'name': 'HMC1',
            'description': None,
            'version': '2.16.0',
            # 'ec-mcl-description':
            # 'network-info':
            # 'machine-info':
            # 'cpc-machine-info':
            'has-hardware-messages': False,
            'hardware-messages': [],
            # 'mobile-app-preferences':
            'sna-name': None,
            'shutdown-in-process': False,
            'shutdown-delay-allowed': False,
            'shutdown-delay-remaining': 0,
            'shutdown-delay-apps': [],
            'shutdown-delay-disable-reasons': [],
            'hma-info': {
                'peer-hmc': None,
            }
        })
        consoles = client.consoles.list(full_properties=True)
        console = consoles[0]

    return console



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`User Pattern` resource represents a pattern for HMC user IDs that are
not defined on the HMC but can be verified by an LDAP server for user
authentication.

User Patterns and user templates allow a system administrator to define a group
of HMC users at once whose user IDs all match a certain pattern (for example,
a regular expression) and who have a certain set of attributes. Each pattern
identifies a template User object which defines many characteristics of such
users. A successful logon with a user ID that matches a User Pattern results in
the creation of a pattern-based user, with many of its attributes coming from
the associated template.

User Patterns are searched in a defined order during logon processing. That
order can be customized through the
:meth:`~zhmcclient.UserPatternManager.reorder` method.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_USER_PATTERN

__all__ = ['UserPatternManager', 'UserPattern']


class UserPatternManager(BaseManager):
    """
    Manager providing access to the :term:`User Pattern` resources of a HMC.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Console` object:

    * :attr:`zhmcclient.Console.user_patterns`

    HMC/SE version requirements:

    * HMC version >= 2.13.0
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #      Console object representing the HMC.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        # Because this resource has case-insensitive names, this list must
        # contain the name property.
        query_props = [
            'name',
            'type',
        ]

        super().__init__(
            resource_class=UserPattern,
            class_name=RC_USER_PATTERN,
            session=console.manager.session,
            parent=console,
            base_uri='/api/console/user-patterns',
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=query_props,
            case_insensitive_names=True)

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: :term:`Console` defining the scope for
        this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the :term:`User Pattern` resources representing the user patterns
        defined in this HMC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * User-related-access permission to the User Pattern objects included
          in the result, or task permission to the "Manage User Patterns" task.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.UserPattern` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'user-patterns'
        list_uri = f'{self.console.uri}/user-patterns'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call
    def create(self, properties):
        """
        Create a new User Pattern in this HMC.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Patterns" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create User Pattern' in the :term:`HMC API` book.

        Returns:

          UserPattern:
            The resource object for the new User Pattern.
            The object will have its 'element-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.console.uri + '/user-patterns',
                                   body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        user_pattern = UserPattern(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return user_pattern

    @logged_api_call
    def reorder(self, user_patterns):
        """
        Reorder the User Patterns of the HMC as specified.

        The order of User Patterns determines the search order during logon
        processing.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Patterns" task.

        Parameters:

          user_patterns (list of :class:`~zhmcclient.UserPattern`):
            The User Patterns in the desired order.

            Must not be `None`.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """  # noqa: E501
        body = {
            'user-pattern-uris': [up.uri for up in user_patterns]
        }
        self.session.post(
            '/api/console/operations/reorder-user-patterns',
            body=body)


class UserPattern(BaseResource):
    """
    Representation of a :term:`User Pattern`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.UserPatternManager`).

    HMC/SE version requirements:

    * HMC version >= 2.13.0
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.UserPatternManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, UserPatternManager), (
            f"Console init: Expected manager type {UserPatternManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this User Pattern.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Patterns" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this UserPattern.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Patterns" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'User Pattern object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)



#!/usr/bin/env python
# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that performs a "Get Partition Properties" operation using the
'properties' query parameter.
"""

import sys
import random
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

print(__doc__)

print(f"Using HMC at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Finding a CPC in DPM mode ...")
    cpcs = client.cpcs.list(filter_args={'dpm-enabled': True})
    if not cpcs:
        print(f"Error: HMC at {host} does not manage any CPCs in DPM mode")
        sys.exit(1)
    cpc = cpcs[0]
    print(f"Using CPC {cpc.name}")

    partitions = cpc.partitions.list()
    partition = random.choice(partitions)
    print(f"Using partition {partition.name} ...")

    print("\nLocal partition properties after list: "
          f"{list(partition.properties.keys())}")

    print("\nExecuting pull_properties() with valid property "
          "'maximum-memory' ...")
    partition.pull_properties('maximum-memory')

    print("Local partition properties after this pull: "
          f"{list(partition.properties.keys())}")

    print("\nExecuting pull_properties() with invalid property "
          "'foo' ...")
    partition.pull_properties('foo')

    print("Local partition properties after this pull: "
          f"{list(partition.properties.keys())}")

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for user roles (on CPCs in DPM mode).

These tests do not change any existing user roles, but create,
modify and delete test user roles.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic UserRole objects (e.g. find_by_name())
UROLE_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in UserRole objects returned by list() without full props
UROLE_LIST_PROPS = ['object-uri', 'name', 'type']

# Properties whose values can change between retrievals of UserRole objects
UROLE_VOLATILE_PROPS = []


def test_urole_find_list(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "user roles")

    # Pick the user roles to test with
    urole_list = console.user_roles.list()
    if not urole_list:
        skip_warn(f"No user roles defined on HMC {hd.host}")
    urole_list = pick_test_resources(urole_list)

    for urole in urole_list:
        print(f"Testing with user role {urole.name!r}")
        runtest_find_list(
            hmc_session, console.user_roles, urole.name, 'name',
            'object-uri', UROLE_VOLATILE_PROPS, UROLE_MINIMAL_PROPS,
            UROLE_LIST_PROPS)


def test_urole_property(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "user roles")

    # Pick the user roles to test with
    urole_list = console.user_roles.list()
    if not urole_list:
        skip_warn(f"No user roles defined on HMC {hd.host}")
    urole_list = pick_test_resources(urole_list)

    for urole in urole_list:
        print(f"Testing with user role {urole.name!r}")

        # Select a property that is not returned by list()
        non_list_prop = 'description'

        runtest_get_properties(urole.manager, non_list_prop)


def test_urole_crud(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a user role.
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console
    hd = hmc_session.hmc_definition

    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC {hd.host} of version {hmc_version} does not yet support "
            "user roles")

    urole_name = TEST_PREFIX + ' test_urole_crud urole1'
    urole_name_new = urole_name + ' new'

    # Ensure a clean starting point for this test
    try:
        urole = console.user_roles.find(name=urole_name)
    except zhmcclient.NotFound:
        pass
    else:
        warnings.warn(
            f"Deleting test user role from previous run: {urole_name!r}",
            UserWarning)
        urole.delete()

    # Test creating the user role

    urole_input_props = {
        'name': urole_name,
        'description': 'Test user role for zhmcclient end2end tests',
    }
    urole_auto_props = {
        'type': 'user-defined',
        'is-inheritance-enabled': False,
    }

    # The code to be tested
    try:
        urole = console.user_roles.create(
            urole_input_props)
    except zhmcclient.HTTPError as exc:
        if exc.http_status == 403 and exc.reason == 1:
            skip_warn(
                f"HMC userid {hd.userid!r} is not authorized for task "
                f"'Manage User Roles' on HMC {hd.host}")
        else:
            raise

    for pn, exp_value in urole_input_props.items():
        assert urole.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"
    urole.pull_full_properties()
    for pn, exp_value in urole_input_props.items():
        assert urole.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"
    for pn, exp_value in urole_auto_props.items():
        assert urole.properties[pn] == exp_value, \
            f"Unexpected value for property {pn!r}"

    # Test updating a property of the user role

    new_desc = "Updated user role description."

    # The code to be tested
    urole.update_properties(dict(description=new_desc))

    assert urole.properties['description'] == new_desc
    urole.pull_full_properties()
    assert urole.properties['description'] == new_desc

    # Test that user roles cannot be renamed

    with pytest.raises(zhmcclient.HTTPError) as exc_info:

        # The code to be tested
        urole.update_properties(dict(name=urole_name_new))

    exc = exc_info.value
    assert exc.http_status == 400
    assert exc.reason == 6
    with pytest.raises(zhmcclient.NotFound):
        console.user_roles.find(name=urole_name_new)

    # Test deleting the user role

    # The code to be tested
    urole.delete()

    with pytest.raises(zhmcclient.NotFound):
        console.user_roles.find(name=urole_name)



# Copyright 2019,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :class:`~zhmcclient.StorageVolumeTemplate` object represents a
:term:`storage volume template` for FCP or FICON (ECKD) that is defined in a
:term:`storage group template`.

Storage volume template objects can be created, updated and deleted.

Storage volume templates are contained in
:term:`storage group templates <storage group template>`.

You can create as many storage volume templates as you want in a storage group
template. When creating a storage group from the storage group template, each
storage volume template will cause a storage volume to be created in the new
storage group.

Storage group templates and storage volume templates only can be defined in
CPCs that are in DPM mode and have SE version >= 2.14.1.
"""


import copy
# from requests.utils import quote

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_STORAGE_TEMPLATE_VOLUME

__all__ = ['StorageVolumeTemplateManager', 'StorageVolumeTemplate']


class StorageVolumeTemplateManager(BaseManager):
    """
    Manager providing access to the
    :term:`storage volume templates <storage volume template>`
    in a particular :term:`storage group template`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.StorageGroupTemplate` object:

    * :attr:`~zhmcclient.StorageGroupTemplate.storage_volume_templates`

    HMC/SE version requirements:

    * SE version >= 2.14.1
    """

    def __init__(self, storage_group_template):
        # This function should not go into the docs.
        # Parameters:
        #   storage_group_template (:class:`~zhmcclient.StorageGroupTemplate`):
        #     Storage group template defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',
            'maximum-size',
            'minimum-size',
            'usage',
        ]

        super().__init__(
            resource_class=StorageVolumeTemplate,
            class_name=RC_STORAGE_TEMPLATE_VOLUME,
            session=storage_group_template.manager.session,
            parent=storage_group_template,
            base_uri=f'{storage_group_template.uri}/storage-template-volumes',
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def storage_group_template(self):
        """
        :class:`~zhmcclient.StorageGroupTemplate`:
        :term:`Storage group template` defining the scope for this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the storage volume templates in this storage group template.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to this storage group template.

        Parameters:

          full_properties (bool):
            Controls that the full set of resource properties for each returned
            storage volume template is being retrieved, vs. only the following
            short set: "element-uri", "name", "size", and "usage".

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.StorageVolumeTemplate` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'storage-template-volumes'
        list_uri = f'{self.storage_group_template.uri}/storage-template-volumes'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call
    def create(self, properties):
        """
        Create a :term:`storage volume template` in this storage group template
        on the HMC.

        This method performs the "Modify Storage Template Properties"
        operation, requesting creation of the volume.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to this storage group template.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          properties (dict): Initial property values for the new volume
            template.

            Allowable properties are the fields defined in the
            "storage-template-volume-request-info" nested object described for
            operation "Modify Storage Template Properties" in the
            :term:`HMC API` book.
            The valid fields are those for the "create" operation. The
            `operation` field must not be provided; it is set automatically
            to the value "create".

            The properties provided in this parameter will be copied and then
            amended with the `operation="create"` field, and then used as a
            single array item for the `storage-template-volumes` field in the
            request body of the "Modify Storage Template Properties" operation.

        Returns:

          StorageVolumeTemplate:
            The resource object for the new storage volume template.
            The object will have the following properties set:

            - 'element-uri' as returned by the HMC
            - 'element-id' as determined from the 'element-uri' property
            - 'class' and 'parent'
            - additional properties as specified in the input properties

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`

        Example::

            stovol1 = fcp_stogrp.storage_volume_templates.create(
                properties=dict(
                    name='vol1',
                    size=30,  # GiB
            ))
        """
        volreq_obj = copy.deepcopy(properties)
        volreq_obj['operation'] = 'create'
        body = {
            'storage-template-volumes': [volreq_obj],
        }
        result = self.session.post(
            self.storage_group_template.uri + '/operations/modify',
            body=body)
        uri = result['element-uris'][0]
        storage_volume_template = self.resource_object(uri, properties)
        # The 'name' property is unique within the parent object. However, it
        # is not returned by this operation, so we don't set the name-to-uri
        # cache. It will be set lazily, upon first use.
        return storage_volume_template


class StorageVolumeTemplate(BaseResource):
    """
    Representation of a :term:`storage volume template`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.StorageVolumeTemplateManager`).

    HMC/SE version requirements:

    * SE version >= 2.14.1
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.StorageVolumeTemplateManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, StorageVolumeTemplateManager), (
            "StorageVolumeTemplate init: Expected manager type "
            f"{StorageVolumeTemplateManager}, got {type(manager)}")
        super().__init__(
            manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this storage volume template on the HMC.

        This method performs the "Modify Storage Template Properties"
        operation, requesting deletion of the volume template.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to the storage group template owning this
          storage volume template.
        * Task permission to the "Configure Storage - System Programmer" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        volreq_obj = {
            'operation': 'delete',
            'element-uri': self.uri,
        }
        body = {
            'storage-template-volumes': [
                volreq_obj
            ],
        }
        self.manager.session.post(
            self.manager.storage_group_template.uri + '/operations/modify',
            resource=self, body=body)

        # pylint: disable=protected-access
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))

        parent_sv_uris = self.manager.parent.get_properties_local(
            'storage-volume-uris')
        if parent_sv_uris:
            try:
                parent_sv_uris.remove(self._uri)
            except ValueError:
                pass

        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this storage volume template on the HMC.

        This method performs the "Modify Storage Template Properties"
        operation, requesting modification of the volume.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.14.1

        Authorization requirements:

        * Object-access permission to the storage group template owning this
          storage volume template.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          properties (dict): New property values for the volume.
            Allowable properties are the fields defined in the
            "storage-template-volume-request-info" nested object for the
            "modify" operation. That nested object is described in operation
            "Modify Storage Template Properties" in the :term:`HMC API` book.

            The properties provided in this parameter will be copied and then
            amended with the `operation="modify"` and `element-uri` properties,
            and then used as a single array item for the
            `storage-template-volumes` field in the request body of the
            "Modify Storage Template Properties" operation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        volreq_obj = copy.deepcopy(properties)
        volreq_obj['operation'] = 'modify'
        volreq_obj['element-uri'] = self.uri
        body = {
            'storage-template-volumes': [volreq_obj],
        }
        self.manager.session.post(
            self.manager.storage_group_template.uri + '/operations/modify',
            resource=self, body=body)
        self.update_properties_local(copy.deepcopy(properties))
